/*
 * Copyright (C) 2018-2021 Texas Instruments Incorporated - https://www.ti.com/
 *
 * SPDX-License-Identifier: Apache-2.0
 */

#ifndef IPC_RPMSG_LINUX_RESOURCE_TABLE_H_
#define IPC_RPMSG_LINUX_RESOURCE_TABLE_H_

#ifdef __cplusplus
extern "C" {
#endif

#include <stddef.h>
#include <zephyr/sys/util_macro.h>

/*
 * Keep structure and defines in the file in sync with the linux kernel
 */

/** \brief Virtio type for Remote Proc Messaging
 */
#define RPMESSAGE_RSC_VIRTIO_ID_RPMSG  (7U)

/**
 * \name Resource Entry type
 *
 * @{
 */

/** \brief trace type */
#define RPMESSAGE_RSC_TYPE_TRACE       (2U)
/** \brief VDEV type */
#define RPMESSAGE_RSC_TYPE_VDEV        (3U)

/**@}*/

/** \brief Macro to specify VRING memory needs to be dynamically allocated by linux */
#define RPMESSAGE_RSC_VRING_ADDR_ANY (0xFFFFFFFFU)

/** \brief Macro to specify linux trace version */
#define RPMESSAGE_RSC_TRACE_INTS_VER0        (0 << 16)
/** \brief Macro to specify linux trace version */
#define RPMESSAGE_RSC_TRACE_INTS_VER1        BIT(16)
/** \brief Max length of trace name */
#define RPMESSAGE_RSC_TRACE_NAME_LEN         (32u)

/**
 * \brief Resource Table Header
 */
struct rpmessage_rschdr {
	uint32_t ver;
	/**< Version Number, set to 1 */
	uint32_t num;
	/**< Number of Entries, MUST be 2 */
	uint32_t reserved[2];
	/**< Reserved for future use, set to 0 */
};

/**
 * \brief Structure used for remoteproc trace
 */
struct rpmessage_rsctrace {
	uint32_t  type;
	/** Type of trace, MUST be set to TYPE_TRACE | TRACE_INTS_VER0 */
	uint32_t  da;
	/**< Device Address, physical address of location of trace buffer in remote side */
	uint32_t  len;
	/**< Length of trace buffer  */
	uint32_t  reserved;
	/**< Reserved for future use, set to 0  */
	uint8_t   name[RPMESSAGE_RSC_TRACE_NAME_LEN];
	/**<  Name of the trace */
};

/**
 * \brief Resource Table Device VRing Structure
 */
struct rpmessage_rscvring {
	uint32_t  da;
	/**< device address, physical address of VRING,
	 *   set to RPMSG_VRING_ADDR_ANY, updated by linux, with actual address
	 */
	uint32_t  align;
	/**< Alignment used between AVAIL and USED structures, updated by linux */
	uint32_t  num;
	/**< Number of message buffers, MUST be 256 */
	uint32_t  notifyid;
	/**< NotifyId for receive channel, set 1 for TX VRING and 2 for RX VRING */
	uint32_t  reserved;
	/**< Reserved for future use, set to 0 */
};

/**
 *  \brief VDEV structure
 */
struct rpmessage_rscvdev {
	uint32_t  type;
	/**< type of VDEV, set to TYPE_VDEV */
	uint32_t  id;
	/**< ID of VDEV, set to VIRTIO_ID_RPMSG */
	uint32_t  notifyid;
	/**< Not used, set to 0  */
	uint32_t  dfeatures;
	/**< Not used, set to 1 */
	uint32_t  gfeatures;
	/**< not used, set to 0 */
	uint32_t  config_len;
	/**< not used, set to 0 */
	uint8_t   status;
	/**< updated by linux, after linux init, this should be 0x7 */
	uint8_t   num_of_vrings;
	/**< number of vrings, set to 2 */
	uint8_t   reserved[2];
	/**< Reserved for future use, set to 0 */
};

/**
 *  \brief IPC Resource Table used by IPC app
 */
struct rpmessage_resourcetable {
	struct rpmessage_rschdr base;
	/**< Header Information */
	uint32_t offset[2];
	/**< offset to VDEV and TRACE entries */
	struct rpmessage_rscvdev vdev;
	/**< VDEV entry */
	struct rpmessage_rscvring  vring0;
	/**< TX VRING  */
	struct rpmessage_rscvring  vring1;
	/**< RX VRING */
	struct rpmessage_rsctrace trace;
	/**< Trace entry  */
};

#ifdef __cplusplus
}
#endif

#endif /* IPC_RPMSG_LINUX_RESOURCE_TABLE_H_ */
